/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

#include "derscrypt.hpp"
#include <string.h>
#include "util.hpp"

using namespace std;

const int BIT_MASKS[]={ 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

sh_ptr<vector<ShUInt> > permute(const vector<ShUInt>& rems, sh_ptr<Key> key,
  bool inverse)
{
 int len=rems.size();
 sh_ptr<vector<ShUInt> > ret(new vector<ShUInt>(len));

 const vector<int>& perm=key->getPermutation(len);
 if (inverse) for (int i=0; i<len; i++) (*ret)[i]=rems[perm[i]];
 else for (int i=0; i<len; i++) (*ret)[perm[i]]=rems[i];

 return ret;
}

sh_ptr<vector<uchar> > permuteBits(const vector<uchar>& bits, sh_ptr<Key> key)
{
 int bitsLen=bits.size();
 sh_ptr<vector<uchar> > ret(new vector<uchar>(bitsLen));

 const vector<int>& perm=key->getPermutation(bitsLen*8);

 for (int i=0, j=0; i<bitsLen; i++) {
     for (int mask=0x80; mask!=0; mask>>=1) {
         int n=perm[j++];
         if ((bits[i]&mask)!=0)
            (*ret)[n/8]|=BIT_MASKS[n%8];
     }
 }

 return ret;
}

ShUInt computeNumber(const vector<ShUInt>& rems, ShUInt b)
{
 ShUInt a=newUInt(*rems[0]);

 for (int i=1, end=rems.size(); i<end; i++) {
     *a*=*b;
     *a+=*rems[i];
 }
 
 return a;
}

ShUInt basicEncrypt(sh_ptr<Key> key, ShUInt a1, bool& substOne)
{
 sh_ptr<vector<ShUInt> > rems=key->computeRemainders(a1);
 sh_ptr<vector<ShUInt> > rems2=permute(*rems, key, false);

 if (*(*rems2)[0]==UInt::zero) {
    *(*rems2)[0]=UInt::one;
    substOne=true;
 }
 else substOne=false;

 ShUInt a2=computeNumber(*rems2, key->getValue());
 return a2;
}

ShUInt basicDecrypt(sh_ptr<Key> key, ShUInt a2, bool substOne)
{
 sh_ptr<vector<ShUInt> > rems=key->computeRemainders(a2);
 
 if (substOne) {
    if (*(*rems)[0]==UInt::one) *(*rems)[0]=UInt::zero;
    else return ShUInt();  // NULL
 }
 
 sh_ptr<vector<ShUInt> > rems2=permute(*rems, key, true);
 ShUInt a1=computeNumber(*rems2, key->getValue());
 
 return a1;
}

sh_ptr<vector<uchar> > appendHash(const vector<uchar>& data,
  const vector<uchar>& hash)
{
 sh_ptr<vector<uchar> > ret(new vector<uchar>);
 ret->reserve(data.size()+hash.size());

 int len=hash.size(), mid=len/2;
 ret->insert(ret->end(), hash.begin(), hash.begin()+len-mid);
 ret->insert(ret->end(), data.begin(), data.end());
 ret->insert(ret->end(), hash.begin()+len-mid, hash.end());

 return ret;
}

void extractHash(const vector<uchar>& bytes, int hashLen,
  sh_ptr<vector<uchar> > data, sh_ptr<vector<uchar> > hash)
{
 data.set(new vector<uchar>);
 data->reserve(bytes.size()-hashLen);
 hash.set(new vector<uchar>);
 hash->reserve(hashLen);

 int mid=hashLen/2;
 hash->insert(hash->end(), bytes.begin(), bytes.begin()+hashLen-mid);
 data->insert(data->end(), bytes.begin()+hashLen-mid, bytes.end()-mid);
 hash->insert(hash->end(), bytes.end()-mid, bytes.end());
}

sh_ptr<vector<uchar> > extractBits(const vector<uchar>& data, int bitsLen)
{
 sh_ptr<vector<uchar> > ret(new vector<uchar>(bitsLen));
 
 for (int i=0, j=1, dataLen=data.size(); i<bitsLen; i++) {
     for (int mask=0x80; mask!=0; mask>>=1) {
         int n=(j++)*(dataLen*8-1)/(bitsLen*8+1);
         if ((data[n/8]&BIT_MASKS[n%8])!=0)
            (*ret)[i]|=mask;
     }
 }

 return ret;
}

sh_ptr<vector<uchar> > computeHash(sh_ptr<Key> key, const vector<uchar>& data,
  const vector<uchar>& prevData)
{
 int len=key->getBytes().size();

 sh_ptr<vector<uchar> > bits;
 if (prevData.size()!=0) bits=extractBits(prevData, len);
 else {
      bits.set(new vector<uchar>(len));
      for (int i=0; i<len; i++) (*bits)[i]=0x55;
 }

 sh_ptr<vector<uchar> > a1buf=appendHash(data, *permuteBits(*bits, key));
 ShUInt a1=newUInt(&(*a1buf)[0], a1buf->size());

 bool substOne;
 ShUInt a2=basicEncrypt(key, a1, substOne);;

 vector<uchar> encr(a2->getSize());
 a2->toBytes(&encr[0]);
 
 sh_ptr<vector<uchar> > hash=permuteBits(*extractBits(encr, len), key);
 if ((*hash)[0]==0) (*hash)[0]=1;

 return hash;
}

sh_ptr<vector<uchar> > blockDecryptImpl(sh_ptr<Key> key,
  const vector<uchar>& encr, const vector<uchar>& prevEncr, bool substOne)
{
 ShUInt aPr2=newUInt(&encr[0], encr.size());
 ShUInt aPr1=basicDecrypt(key, aPr2, substOne);
 if (!aPr1.get()) return sh_ptr<vector<uchar> >();  // NULL

 vector<uchar> decr(aPr1->getSize());
 aPr1->toBytes(&decr[0]);

 sh_ptr<vector<uchar> > plain, hash;
 extractHash(decr, key->getBytes().size(), plain, hash);

 sh_ptr<vector<uchar> > hash2=computeHash(key, *plain, prevEncr);
 if (*hash!=*hash2) return sh_ptr<vector<uchar> >();  // NULL

 return plain;
}  

sh_ptr<vector<uchar> > blockEncrypt(sh_ptr<Key> key, const vector<uchar>& plain,
  const vector<uchar>& prevEncr)
{
 sh_ptr<vector<uchar> > hash=computeHash(key, plain, prevEncr);

 sh_ptr<vector<uchar> > aPr1buf=appendHash(plain, *hash);
 ShUInt aPr1=newUInt(&(*aPr1buf)[0], aPr1buf->size());

 bool substOne;
 ShUInt aPr2=basicEncrypt(key, aPr1, substOne);
 
 sh_ptr<vector<uchar> > encr(new vector<uchar>(aPr2->getSize()));
 aPr2->toBytes(&(*encr)[0]);

 if (substOne && blockDecryptImpl(key, *encr, prevEncr, false).get())
    error("Cannot encrypt this data using this key: Hash collision\n");

 return encr;
}

sh_ptr<vector<uchar> > blockDecrypt(sh_ptr<Key> key, const vector<uchar>& encr,
  const vector<uchar>& prevEncr)
{
 sh_ptr<vector<uchar> > decr=blockDecryptImpl(key, encr, prevEncr, false);
 if (decr.get()) return decr;

 decr=blockDecryptImpl(key, encr, prevEncr, true);
 if (!decr.get()) error("Incorrect key and/or data: Hash mismatch\n");

 return decr;
}

void read(FILE* in, int len, vector<uchar>& buf)
{
 buf.resize(len);
 buf.resize(fread(&buf[0], 1, len, in));
 if (ferror(in)) error("File read error\n");
}

int readInt16(FILE* f)
{
 int ch1=fgetc(f);
 if (ch1==EOF) return EOF;

 int ch2=fgetc(f);
 if (ch2==EOF) error("Invalid data: Invalid block format\n");

 return (ch1<<8)+ch2;
}

void write(FILE* out, const vector<uchar>& buf)
{
 fwrite(&buf[0], buf.size(), 1, out);
 if (ferror(out)) error("File write error\n");
}

void writeInt16(FILE* f, int val)
{
 fputc(val>>8, f);
 fputc(val&0xFF, f);
}

void checkKey(const void* ptr, int len)
{
 const uchar* key=(uchar*)ptr;
 
 if (len<=0)
    error("Invalid key: invalid key length\n");
 if (key[0]==0)
    error("Suspectable key: Leading zeros are useless\n");
 if ((key[len-1]&1)==0)
    error("Suspectable key: Even values are not recommended\n");
}

int minBlockLength(int keyLength)
{
 if (keyLength<16) keyLength=16;       // 128->40
 else if (keyLength>64) keyLength=64;  // 512->100

 int n=40+((keyLength-16)*5+2)/4;
 return keyLength*n;
}

int streamEncrypt(sh_ptr<Key> key, FILE* in, FILE* out)
{
 vector<uchar> bufs[2];

 int minLen=minBlockLength(key->getBytes().size());
 read(in, minLen, bufs[false]);
 
 int readCnt=bufs[false].size();
 if (readCnt!=minLen)
    error("Insufficient data length: At least %d bytes are required\n", minLen);

 sh_ptr<vector<uchar> > prevEncr(new vector<uchar>);
 for (bool n=true; int(bufs[!n].size())==minLen; n=!n) {
     read(in, minLen, bufs[n]);
     readCnt+=bufs[n].size();

     if (int(bufs[n].size())<minLen)
        bufs[!n].insert(bufs[!n].end(), bufs[n].begin(), bufs[n].end());

     sh_ptr<vector<uchar> > encr=blockEncrypt(key, bufs[!n], *prevEncr);
     prevEncr=encr;

     writeInt16(out, encr->size());
     write(out, *encr);
 }

 if (fflush(out)==EOF) error("File write error\n");
 
 return readCnt;
}

int streamDecrypt(sh_ptr<Key> key, FILE* in, FILE* out)
{
 int readCnt=0, minLen=minBlockLength(key->getBytes().size());

 sh_ptr<vector<uchar> > encr(new vector<uchar>), prevEncr(new vector<uchar>);
 for (int len; (len=readInt16(in))!=EOF; ) {
     readCnt+=2;
     if (len<=minLen || len>=minLen*3)
        error("Invalid data: Invalid block format\n");

     read(in, len, *encr);
     readCnt+=encr->size();

     if (int(encr->size())!=len) error("Invalid data: Invalid block format\n");

     sh_ptr<vector<uchar> > decr=blockDecrypt(key, *encr, *prevEncr);
     prevEncr.swap(encr);

     write(out, *decr);
 }
 
 if (fflush(out)==EOF) error("File write error\n");

 return readCnt;
}

