/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Provides command line UI to DersCrypt cipher.  
 */

#include <stdlib.h>
#include <time.h>
#include <string>
#include <exception>
#include "derscrypt.hpp"

using namespace std;

void invalidArgs()
{
 fprintf(stderr,
   "Invalid args.\n"
   "Must be: e|d key_file input_file output_file\n"
 );
 exit(1);
}

sh_ptr<vector<uchar> > readFile(const string& name)
{
 sh_ptr<vector<uchar> > ret(new vector<uchar>);
 ret->reserve(64);

 FILE* f=fopen(name.c_str(), "rb");
 if (!f) error("Can't open \"%s\" file\n", name.c_str());

 for (int ch; (ch=fgetc(f))!=EOF; )
     ret->push_back(ch);

 fclose(f);

 return ret;
}

int main(int argc, char* argv[])
{
 string errMsg;
 try {
     printf(
       "DersCrypt v1.1 (2005-12-24), symmetric-key block cipher, freeware, "
       "CPU=%s\n"
       "Copyright (C) Sergey P. Derevyago\n\n", UInt::cpuString()
     );

     if (argc!=5) invalidArgs();

     string sCmd=argv[1], sKey=argv[2], sIn=argv[3], sOut=argv[4];
     if ( !(sCmd=="e" || sCmd=="d") ) invalidArgs();

     sh_ptr<vector<uchar> > kbuf=readFile(sKey);
     sh_ptr<Key> key=newKey(&(*kbuf)[0], kbuf->size());

     FILE* fin=fopen(sIn.c_str(), "rb");
     if (!fin) error("Can't open \"%s\" file\n", sIn.c_str());

     FILE* fout=fopen(sOut.c_str(), "wb");
     if (!fout) {
        fclose(fin);
        error("Can't open \"%s\" file\n", sOut.c_str());
     }

     time_t start=time(0);

     int cnt= (sCmd=="e") ?
       streamEncrypt(key, fin, fout) : streamDecrypt(key, fin, fout);

     int sec=int(time(0)-start);
     if (sec) printf("%d sec, %d b/s\n", sec, cnt/sec);

     fclose(fout);
     fclose(fin);
 }
 catch (const exception& e) {
       errMsg="std::exception caught: ";
       errMsg+=e.what();
 }
 catch (...) {
       errMsg="Unknown exception caught";
 }
 if (!errMsg.empty()) error("%s\n", errMsg.c_str());

 return 0;
}

